import java.lang.*;
import java.io.*;
import java.util.*;
import java.awt.*;

/*------------------------------------------
StrIntArr
  Maps an ID3 class name to a count.
  The TotalWordMap is a hashtable of
  StrIntArr objects where a word is mapped
  to a StrIntArr.
  ------------------------------------------*/
class StrIntArr extends Hashtable
{
    // Add a delta to # of times a class contains this word
	public void AddToClassCnt( String s, int i )
	{
		if ( containsKey( s ) )
		{
			Integer IV = (Integer)get( s );
			remove( s );
			int iNew = IV.intValue() + i;
			if ( iNew > 0 )
    			put( s, new Integer( iNew ) );
		} else {
		    if ( i > 0 )
    			put( s, new Integer( i ) );
		}
	}

    // Return the # of times a given class contains this word
	public int getClassCnt( String s )
	{
		if ( containsKey( s ) )
		{
			Integer I = (Integer)get( s );
			return I.intValue();

		} else {
			return 0;

		}
	}


}


/*------------------------------------------
TreeNode
   This class is a node on a decision tree.
  ------------------------------------------*/
class TreeNode {

    // The class that this node predicts
    String      sPredictedClass;

    //  True if a leaf (Left & Right will be null)
    boolean     bIsLeaf;

    // Branches from this node
    TreeNode    Left;
    TreeNode    Right;

    // The string that the node tests for
    String      sTest;

    // false if testing for word's absence
    boolean     bPosTest;

    // Count of the number of examples classified by this node
    // together with the number of known misclassified cases.
    int         iNumCases;
    int         iNumIncorrect;

    // Constructor for building a leaf
    TreeNode( String sClass, int iNumCases, int iNumIncorrect )
    {
        sPredictedClass = sClass;
        bIsLeaf = true;
        Left = null;
        Right = null;
        sTest = "";
        bPosTest = true;
        this.iNumCases = iNumCases;
        this.iNumIncorrect = iNumIncorrect;
    }

    // Constructor for building a non-leaf
    TreeNode( String sClass, String sTest, boolean bPosTest )
    {
        sPredictedClass = sClass;
        bIsLeaf = false;
        Left = null;
        Right = null;
        this.sTest = sTest;
        this.bPosTest = bPosTest;
        iNumCases = 0;
        iNumIncorrect = 0;

    }


    // Returns the actual number of cases of
    //    a given class covered by this subtree
    int getActualCnt( String sClass )
    {
        if ( bIsLeaf )
        {
            if ( sClass.equalsIgnoreCase( sPredictedClass ) )
                return iNumCases - iNumIncorrect;
            else
                return iNumIncorrect;
        }
        return Left.getActualCnt( sClass ) + Right.getActualCnt( sClass );
    }

    // Natural logarithm
    static double LN( double x )
    {
        return java.lang.Math.log( x ) / java.lang.Math.log( isE );
    }

    /*
      Statistic data.
     */
    static double Z_Squared = 0.451;     // 25% data either side of the mean
    static double dbBinomialRate = 0.25;
    static double isE = 2.7183;

    static double calcPredictedErrs( double N, double e )
    {
        if ( e == 0 )
        {
            // Simple binomial rate.
            double dbPredictedRate = 1 - java.lang.Math.exp( LN(dbBinomialRate) / N );
            return N * dbPredictedRate;
        }

        // Estimate error rate with normal approximation to binomial
        double dbPredictedRate;

        dbPredictedRate =
          (e + 0.5 + Z_Squared/2 + java.lang.Math.sqrt( Z_Squared * ((e+0.5)*(1-(e+0.5)/N)
            + Z_Squared/4)) ) / (N + Z_Squared);
        return N * dbPredictedRate;

    }

    // Return the # of predicted errors for this subtree (or leaf)
    double getPredictedErrs()
    {
        if ( bIsLeaf )
        {
            return calcPredictedErrs( iNumCases, iNumIncorrect );
        }

        return Left.getPredictedErrs() + Right.getPredictedErrs();
    }


    // Print the tree into a TextArea object
    void Print( TextArea ta )
    {
        Print( ta, 0 );
    }

    // Adds whitespace to get to the i'th tab stop
    void SetTab( int i, TextArea ta )
    {
        while( i > 0 )
        {
            ta.appendText( "  " );
            i = i -1;
        }
    }

    // Print the tree at a given indention level
    void Print( TextArea ta, int iLevel )
    {
        SetTab( iLevel, ta );
        if ( bIsLeaf )
        {
            ta.appendText(
              "CLASS: " + sPredictedClass + " (" + iNumCases + "/" + iNumIncorrect + ")\n"
            );
        }
        else {
            if ( bPosTest )
            {
                ta.appendText( "If " + sTest + " is an element.\n" );
                Left.Print( ta,  iLevel+1);

                SetTab( iLevel, ta );
                ta.appendText( "else If " + sTest + " is not an element.\n" );
                Right.Print( ta, iLevel+1);
            } else {
                ta.appendText( "If " + sTest + " is NOT an element.\n" );
                Left.Print( ta,  iLevel+1);

                SetTab( iLevel, ta );
                ta.appendText( "else If " + sTest + " is an element.\n" );
                Right.Print( ta, iLevel+1);
            }
        }
    }
}


/*------------------------------------------
TotalWordMap
  Utility class that maps words to classes.
  The hashtable uses words as key values.
  The value of the key is a StrIntArr that
  describes how many times the word is used
  in a given class.
  ------------------------------------------*/
class TotalWordMap extends Hashtable
{
    // Add a BagOWords to the map
    public void addBOW( BagOWords bow )
    {
        Enumeration e = bow.enumWords();
        while( e.hasMoreElements() )
        {
            String s = (String)e.nextElement();
            if ( !containsKey( s ) )
                put( s, new StrIntArr() );

            StrIntArr   Classes = (StrIntArr)get( s );
            Classes.AddToClassCnt( bow.sClass, 1 );
        }
    }

    // How many times does word sWord appear in a class of sClass
    public int getClassCnt( String sWord, String sClass )
    {
        StrIntArr  Classes = (StrIntArr)get( sWord );
        return Classes.getClassCnt( sClass );
    }

    // How many times does word sWord appear in a class of !sClass
    public int getNotClassCnt( String sWord, String sClass )
    {
        StrIntArr  Classes = (StrIntArr)get( sWord );

        int i = 0;
        Enumeration e = Classes.keys();
        while( e.hasMoreElements() )
        {
            String s = (String)e.nextElement();
            i = i + Classes.getClassCnt( s );
        }
        return i - getClassCnt( sWord, sClass );
    }

    // Fixup the word counts by removing a BagOWords from map
    public void fixupRemove( BagOWords bow )
    {
        Enumeration e = bow.enumWords();
        while( e.hasMoreElements() )
        {
            String  s = (String)e.nextElement();
            StrIntArr  Classes = (StrIntArr)get( s );
            Classes.AddToClassCnt( bow.sClass, -1 );
        }

    }

    // Remove all references to a certain word
    public void deleteWord( String sWord )
    {
        remove( sWord );
    }

    // Enumerate the words in the map
    public Enumeration enumWords()
    {
        return keys();
    }
}





/*------------------------------------------
ExampleSet
  Read in a set of examples and apply
  the ID3 and partial C4.5 algorithm to
  those samples.  Print the results to
  a TextArea control.
  ------------------------------------------*/
class ExampleSet
{
    // Base 10 Logarithm of 2.0
    static  double dbLog2 = java.lang.Math.log( 2.0 );

    // Smallest suggested size of a single leaf node
    static int  iSmallestNodeSize = 1;

    // Minimum number of examples needed to do a test
    static int  iMinExamples = 10;

    // Calculate the entropy of an array of values
    static public double Entropy( double args[] )
    {
        double db = 0.0;
        int     i = 0;

        while( i < args.length )
        {
            if ( args[i] > 0 )
                db = db + args[i] * java.lang.Math.log( args[i] ) / dbLog2;
            i = i +1;
        }
        db = -db;
        return db;
    }

    // Calculate the entropy of two values
    static public double Entropy( double x, double y )
    {
        double args[] = new double[2];
        args[0] = x;
        args[1] = y;
        return Entropy( args );
    }

    /* Calculate conditional entropy
        Arguments are from this matrix:
                     Passed      Failed
                     a test      a test
                -------------------------
        Class A|       p           pp
        Class B|       n           np
    */
    static  public double CondEntropy( int p, int n, int pp, int np )
    {
        int iTotal = p+n+pp+np;

        double  dbSubTotal1 = p+n;
        double  dbSubTotal2 = (double)iTotal-dbSubTotal1;
        double  dbPortion1 = dbSubTotal1 / ( iTotal );
        double  dbPortion2 = 1.0 - dbPortion1;


        double db = 0;
        if ( dbPortion1 > 0.0 )
            db = db + dbPortion1 * Entropy( p/dbSubTotal1, n/dbSubTotal1 );
        if ( dbPortion2 > 0.0 )
            db = db + dbPortion2 * Entropy( pp/dbSubTotal2, np/dbSubTotal2 ) ;
        return db;
    }

    // Read the next example from a stream and return BagOWords
    static BagOWords ReadOneExample( StreamTokenizer stk )
    {

        // Save the Class.
        BagOWords bow = new BagOWords( stk.sval );

        try {

            // Read in the data
            stk.nextToken();
            while( stk.ttype != StreamTokenizer.TT_EOF )
            {
                if ( stk.ttype == StreamTokenizer.TT_WORD )
                {
                    String  s = stk.sval;
                    if ( s.equalsIgnoreCase( "ENDOFEXAMPLE" ) )
                    {
                        stk.nextToken();
                        break;
                    }
                    bow.addWord( s );
                    //bowTotal.addWord( s );
                }
                stk.nextToken();
            }
        } catch ( Exception e )
        {}
        return bow;

    }

    // Initialization method.
    void init( File f, TextArea ta )
    {
        if ( !f.exists() )
            return;

        try {

            FileInputStream    fis = new FileInputStream( f );
            BufferedInputStream bis = new BufferedInputStream( fis, 8192 );
            StreamTokenizer     stk = new StreamTokenizer( bis );

            stk.lowerCaseMode( true );
            stk.slashSlashComments( true );
            stk.slashStarComments( true );
            stk.quoteChar( '"' );
            stk.wordChars( 33, 127 );

            Vector        vExamples = new Vector();
            TotalWordMap  twm       = new TotalWordMap();
            StrIntArr     aClassCnt = new StrIntArr();

            stk.nextToken();
            if ( stk.ttype != StreamTokenizer.TT_WORD )
                return;

            String sMainClass = stk.sval;
            stk.nextToken();
            String sNotMainClass = stk.sval;
            stk.nextToken();

            int i = 0;
            while( stk.ttype != StreamTokenizer.TT_EOF )
            {
                BagOWords bow = ReadOneExample( stk );
                twm.addBOW( bow );
                aClassCnt.AddToClassCnt( bow.sClass, 1 );
                vExamples.addElement( bow );

                i = i +1;
                if ( i % 10 == 0 )
                {
                    ta.appendText( "Parsed example " + i + "\n" );
                }

            }


            fis.close();

            ta.appendText( "Building decision tree.\n" );

            TreeNode tn;
            tn = ID3( vExamples, twm, sMainClass, sNotMainClass );
            ta.appendText("===== INITIAL TREE =======\n" );
            tn.Print( ta );

            ta.appendText( "===== FINAL TREE =======\n" );

            tn = PruneTree( tn, sMainClass, sNotMainClass );
            tn.Print( ta );


        } catch( Exception e )
        {}

    }


    /* Calculate maximum gain test (C4.5) from list of examples,
        TotalWordMap, and two class names.
     */
    static TreeNode  MaxGainTest( Vector vExamples, TotalWordMap twm, String sClassA, String sClassB )
    {
        int         iTotalMain = 0;
        int         iTotalNotMain = 0;
        boolean     bPosTest = true;

        Enumeration e = vExamples.elements();
        while ( e.hasMoreElements() )
        {
            BagOWords   bow = (BagOWords)e.nextElement();
            if ( bow.sClass.equalsIgnoreCase( sClassA ) )
            {
                iTotalMain = iTotalMain +1;
            }
            else
            {
                iTotalNotMain = iTotalNotMain +1;
            }
        }

        if ( iTotalMain<iSmallestNodeSize |
             iTotalNotMain<iSmallestNodeSize |
             (iTotalMain + iTotalNotMain < iMinExamples ) )
        {
            int iTotal = iTotalMain + iTotalNotMain;
            if ( iTotalMain < iTotalNotMain )
                return new TreeNode( sClassB, iTotal, iTotalMain );
            else
                return new TreeNode( sClassA, iTotal, iTotalNotMain );
        }

        double  dbBestGain = -1.0;
        double  dbBaseEntropy = Entropy( (double)iTotalMain/(iTotalMain+iTotalNotMain), (double)iTotalNotMain/(iTotalMain+iTotalNotMain) );

        e = twm.enumWords();

        String sBestWord = "";
        int     pBest=0, nBest=0;

        while ( e.hasMoreElements() )
        {
            String  sWord = (String)e.nextElement();
            int     p=0, n=0;

            // Number of true positives
            p = twm.getClassCnt( sWord, sClassA );

            // Number of false positives
            n = twm.getNotClassCnt( sWord, sClassA );

            double dbEntropy = CondEntropy( p, n, iTotalMain-p, iTotalNotMain-n );
            double dbGain = dbBaseEntropy - dbEntropy;

            if ( dbGain > dbBestGain )
            {
                dbBestGain = dbGain;
                sBestWord = sWord;
                pBest = p;
                nBest = n;
                bPosTest = true;
                if ( p<n )
                {
                    pBest = iTotalMain-p;
                    nBest = iTotalNotMain-n;
                    bPosTest = false;
                }
            }

        }

        int iTotal = iTotalMain + iTotalNotMain;
        boolean bIsLeaf = false;

        if ( dbBestGain <= 0 )
            bIsLeaf = true;

        if (pBest==0 && nBest==iTotal )
            bIsLeaf = true;

        if ( nBest==0 && pBest==iTotal )
            bIsLeaf = true;

        if ( bIsLeaf )
        {
            if ( iTotalMain < iTotalNotMain )
                return new TreeNode( sClassB, iTotal, iTotalMain );
            else
                return new TreeNode( sClassA, iTotal, iTotalNotMain );
        }

        return new TreeNode( sClassA, sBestWord, bPosTest );


    }


    // ID3/(subset C4.5) algorithm
    TreeNode   ID3( Vector vExamples, TotalWordMap twm,
                   String sMainClass, String sNotMainClass )
    {
        Vector       vLeft = new Vector();
        TotalWordMap twmLeft = new TotalWordMap();
        String       sRuleLeft, sRuleRight;

        TreeNode tn = MaxGainTest( vExamples, twm,
	                          sMainClass, sNotMainClass );
        if ( tn.bIsLeaf )
            return tn;

        // Build the left branch data structures
        Enumeration e = vExamples.elements();
        while ( e.hasMoreElements() )
        {
            BagOWords  b = (BagOWords)e.nextElement();
            if ( b.hasWord( tn.sTest ) == tn.bPosTest )
            {
                twmLeft.addBOW( b );
                vLeft.addElement( b );
            }
        }


        // Delete the word from all examples
        e = vExamples.elements();
        while( e.hasMoreElements() )
        {
            BagOWords b = (BagOWords)e.nextElement();
            if ( b.hasWord( tn.sTest ) )
                b.removeWord( tn.sTest );
        }
        twm.deleteWord( tn.sTest );
        twmLeft.deleteWord( tn.sTest );

        // Remove examples from the right branch data
        e = vLeft.elements();
        while( e.hasMoreElements() )
        {
            BagOWords b = (BagOWords)e.nextElement();
            vExamples.removeElement( b );
            twm.fixupRemove( b );
        }

        // Recursively build the left and right branches.
        tn.Left = ID3( vLeft, twmLeft, sMainClass, sNotMainClass );
        tn.Right = ID3( vExamples, twm, sMainClass, sNotMainClass );
        return tn;

    }


    // Prune a decision tree
    TreeNode PruneTree( TreeNode tn, String sClassA, String sClassB )
    {
        if ( tn.bIsLeaf )
            return tn;

        // Prune lower levels of tree first
        if ( !tn.Left.bIsLeaf )
            tn.Left = PruneTree( tn.Left, sClassA, sClassB );

        if ( !tn.Right.bIsLeaf )
            tn.Right = PruneTree( tn.Right, sClassA, sClassB );

        int iActualA = tn.getActualCnt( sClassA );
        int iActualB = tn.getActualCnt( sClassB );
        int iTotalOverall = iActualA + iActualB;

        double dbNumErrLeafA = tn.calcPredictedErrs( iTotalOverall, iActualB );
        double dbNumErrLeafB = tn.calcPredictedErrs( iTotalOverall, iActualA );
        double dbTreeRate = tn.getPredictedErrs();

        // Replace with a leaf of type "A"?
        if ( dbNumErrLeafA <= dbNumErrLeafB && dbNumErrLeafA <= dbTreeRate )
            return new TreeNode( sClassA, iTotalOverall, iActualB );

        // Replace with a leaf of type "B"?
        if ( dbNumErrLeafB <= dbTreeRate )
            return new TreeNode( sClassB, iTotalOverall, iActualA );

        return tn;

    }


}

